import logging as logger
import os

from obs import ObsClient
from tenacity import retry, stop_after_attempt, wait_fixed

from ..exception.obs_exception import OBSException, OBSUploadException, \
    OBSDownloadException

LOGGER = logger.getLogger('obs')


class OBS:
    """
        OBS SDK Operation using temporary AK/SK
    """

    def __init__(self, server=None, ak=None, sk=None, security_token=None,
                 is_secure=True, path_style=True):
        """AK SK and security_token comes from IAM's temporary AK/SK API
        """
        if ak is None or sk is None or server is None:
            raise Exception("Parameter is illegal!")

        self.server = server
        self.ak = ak
        self.sk = sk
        if security_token and security_token.strip() != "":
            self.security_token = security_token
        else:
            self.security_token = None

        self.is_secure = is_secure
        self.path_style = path_style
        self.obs_client = ObsClient(access_key_id=self.ak,
                                    secret_access_key=self.sk,
                                    security_token=self.security_token,
                                    is_secure=self.is_secure,
                                    server=self.server,
                                    path_style=self.path_style)

    def get_obs_client(self):
        return self.obs_client

    def check_bucket_path(self, bucket_path):
        if bucket_path.startswith('/'):
            bucket_path = bucket_path[1:]
        if bucket_path.endswith('/'):
            bucket_path = bucket_path[:-1]
        return bucket_path

    def list_buckets(self):
        try:
            resp = self.obs_client.listBuckets()
            if resp.status > 300:
                raise OBSException(code=resp.errorCode,
                                   message=resp.errorMessage)
            else:
                bucket_list = []
                buckets = resp.body.buckets
                for bucket in buckets:
                    bucket_list.append(bucket['name'])
                return bucket_list
        except OBSException:
            raise OBSException(code=resp.errorCode, message=resp.errorMessage)
        except Exception as e:
            raise Exception("Create OBS bucket failed! ", e)

    @retry(
        wait=wait_fixed(1),
        stop=stop_after_attempt(3))
    def list_objects(
            self,
            bucket_name,
            prefix=None,
            marker=None,
            max_keys=None,
            delimiter=None,
            extension_headers=None):
        """
        list objects , support retry
        :param bucket_name:
        :param prefix:
        :param marker:
        :param max_keys:
        :param delimiter:
        :param extension_headers:
        :return:
        """
        return self.obs_client.listObjects(
            bucketName=bucket_name,
            prefix=prefix,
            marker=marker,
            max_keys=max_keys,
            delimiter=delimiter,
            extensionHeaders=extension_headers)

    def create_bucket(self, bucket_name, location):
        bucket_list = self.list_buckets()
        try:
            if bucket_name in bucket_list:
                print("%s is existed" % bucket_name)
                return bucket_name
            else:
                resp = self.obs_client.createBucket(bucketName=bucket_name,
                                                    location=location)
                if resp.status > 300:
                    raise OBSException(code=resp.errorCode,
                                       message=resp.errorMessage)
                else:
                    print("Successfully create bucket %s " % bucket_name)
                    return bucket_name
        except OBSException:
            raise OBSException(code=resp.errorCode, message=resp.errorMessage)
        except Exception as e:
            raise Exception("Create OBS bucket failed! ", e)

    def delete_bucket(self, bucket_name):
        try:
            resp = self.obs_client.deleteBucket(bucketName=bucket_name)
            if resp.status > 300:
                raise OBSException(code=resp.errorCode,
                                   message=resp.errorMessage)
            else:
                print("Successfully delete bucket %s " % bucket_name)
                return resp
        except OBSException:
            raise OBSException(code=resp.errorCode, message=resp.errorMessage)
        except Exception as e:
            raise Exception("Delete OBS bucket failed! ", e)

    def list_top_num_objects(self, bucket_name, num):
        """You can get obejects using following example:
               for content in resp.body.contents:
                   print('\t' + content.key + ' etag[' + content.etag + ']')
        """
        try:
            resp = self.obs_client.listObjects(bucket_name, max_keys=num)
            if resp.status > 300:
                raise OBSException(code=resp.errorCode,
                                   message=resp.errorMessage)
            else:
                print("Successfully list bucket %s object" % bucket_name)
                return resp
        except OBSException:
            raise OBSException(code=resp.errorCode, message=resp.errorMessage)
        except Exception as e:
            raise Exception("List OBS objects failed! ", e)

    def put_object(self, bucket_path, local_file_path):
        """put local file to OBS
           if the bucket_name does not exist in OBS, then it will throw 404 exception
           if the bucket path does not exist in the bucket Name, then it will be created
           if OBS has existed the file in the specified path, then it will be overwritten
        """
        try:
            if not os.path.exists(local_file_path):
                raise Exception("File " + local_file_path + " does not exist!")
            file_name = os.path.split(local_file_path)[-1]
            resp = self.obs_client.putFile(bucket_path, file_name,
                                           local_file_path)
            if resp.status > 300:
                raise OBSUploadException(code=resp.errorCode,
                                         message=resp.errorMessage)
            else:
                # print("Successfully upload file %s to OBS %s" % (local_file_path, bucket_path))
                LOGGER.info(f"Successfully upload file {local_file_path} to OBS {bucket_path}")
        except OBSUploadException:
            raise OBSException(code=resp.errorCode, message=resp.errorMessage)
        except Exception as e:
            raise Exception("Upload file to OBS failed! ", e)

    @retry(
        wait=wait_fixed(1),
        stop=stop_after_attempt(3))
    def put_file(
            self,
            bucket_name,
            object_key,
            content,
            metadata=None,
            headers=None,
            progress_callback=None,
            auto_close=True,
            extension_headers=None
    ):
        """
        put object, support retry
        :param bucket_name:
        :param object_key:
        :param content:
        :param metadata:
        :param headers:
        :param progress_callback:
        :param auto_close:
        :param extension_headers:
        :return:
        """
        return self.obs_client.putObject(
            bucketName=bucket_name,
            objectKey=object_key,
            content=content,
            metadata=metadata,
            headers=headers,
            progressCallback=progress_callback,
            autoClose=auto_close,
            extensionHeaders=extension_headers
        )

    @retry(
        wait=wait_fixed(1),
        stop=stop_after_attempt(3))
    def append_object(
            self,
            bucket_name,
            object_key,
            content=None,
            metadata=None,
            headers=None,
            progress_callback=None,
            auto_close=True,
            extension_headers=None):
        """
        append object, support retry
        :param bucket_name:
        :param object_key:
        :param content:
        :param metadata:
        :param headers:
        :param progress_callback:
        :param auto_close:
        :param extension_headers:
        :return:
        """
        return self.obs_client.appendObject(
            bucketName=bucket_name,
            objectKey=object_key,
            content=content,
            metadata=metadata,
            headers=headers,
            progressCallback=progress_callback,
            autoClose=auto_close,
            extensionHeaders=extension_headers)

    def put_directory(self, bucket_path, local_directory_path):
        """put multi local files to OBS
           if the bucket_name does not exist in OBS, then it will throw 404 exception
           if the bucket path does not exist in the bucket Name, then it will be created
           if OBS has existed the file in the specified path, then it will be overwritten
        """
        try:
            local_directory = local_directory_path.rstrip('/')
            if not os.path.exists(local_directory):
                raise Exception(
                    "Directory " + local_directory + " does not exist!")
            file_name = os.path.split(local_directory)[-1]
            resp = self.obs_client.putFile(bucket_path, file_name,
                                           local_directory)
            for file in resp:
                if os.path.isdir(
                        os.path.split(local_directory)[0] + '/' + file[0]):
                    self.put_directory(
                        bucket_path + '/' + file[0][:file[0].rfind('/')],
                        os.path.split(local_directory)[0] + '/' + file[0])
                else:
                    if file[1].status > 300:
                        raise Exception(file[1].reason)
                    else:
                        # print("Successfully upload file %s to OBS %s" % (local_directory, bucket_path))
                        LOGGER.info(f"Successfully upload file {local_directory} to OBS {bucket_path}")
        except Exception as e:
            raise Exception("Upload file to OBS failed, {}".format(e))

    def put_multi_objects(self, bucket_path, local_file_paths):
        """put local directory to OBS, splitted using ','
           if the bucket_name does not exist in OBS, then it will throw 404 exception
           if the bucket path does not exist in the bucket Name, then it will be created
           if OBS has existed the file in the specified path, then it will be overwritten
        """
        for local_file_path in local_file_paths:
            try:
                if not os.path.exists(local_file_path):
                    raise Exception(
                        "File " + local_file_path + " does not exist!")
                if os.path.isdir(local_file_path):
                    self.put_directory(bucket_path, local_file_path)
                else:
                    self.put_object(bucket_path, local_file_path)
            except Exception as e:
                raise Exception("Upload multi files to OBS failed! ", e)

    def download_object(self, bucket_path, local_file_path):
        """download OBS file to your local file
           if local path has existed the file, then it will be overwritten
        """
        try:
            bucket_path_split = bucket_path.split('/', 1)
            bucket_name = bucket_path_split[0]
            object_name = bucket_path_split[1]
            resp = self.obs_client.getObject(bucket_name, object_name,
                                             downloadPath=local_file_path)
            if resp.status > 300:
                raise OBSDownloadException(code=resp.errorCode,
                                           message=resp.errorMessage)
            else:
                LOGGER.info(f"Successfully download file {bucket_path} from OBS to local {local_file_path}")
        except OBSDownloadException:
            raise OBSDownloadException(code=resp.errorCode,
                                       message=resp.errorMessage)
        except Exception as e:
            raise Exception("Download file from OBS failed! ", e)

    def download_directory(self, bucket_path, local_storage_path):
        """download OBS directory files to your local path
           if local path has existed the file, then it will be overwritten
        """
        try:
            base_dir = local_storage_path
            bucket_path_split = bucket_path.split('/', 1)
            bucket_name = bucket_path_split[0]
            object_name = bucket_path_split[1] + "/"

            obs_objects_lists = self._list_obs_objects(bucket_name,
                                                       object_name)

            for list_objects_resp in obs_objects_lists:
                for content in list_objects_resp.body.contents:
                    if content.key[-1] != '/':
                        bucket_path = bucket_name + "/" + content.key
                        if '/' in content.key.split(object_name)[-1]:
                            local_storage_dir = base_dir + '/' + \
                                                content.key.split(object_name)[
                                                    -1].rsplit('/', 1)[
                                                    0]
                        else:
                            local_storage_dir = base_dir
                        os.makedirs(local_storage_dir, exist_ok=True)
                        local_file_path = local_storage_dir.rstrip('/') + '/' + \
                                          content.key.rsplit('/', 1)[-1]
                        self.download_object(bucket_path, local_file_path)
                    else:
                        local_storage_dir = base_dir + '/' + \
                                            content.key.split(object_name)[-1]
                        os.makedirs(local_storage_dir.rstrip('/'), exist_ok=True)
        except Exception as e:
            raise Exception("Download file from OBS failed! ", e)

    def get_object_size(self, bucket_name, object_name):
        try:
            resp = self.obs_client.getObjectMetadata(bucket_name, object_name)
            if resp.status > 300:
                raise OBSException(code=resp.errorCode,
                                   message=resp.errorMessage)
            header = dict(resp.header)
            object_size = int(header.get('content-length'))
            return object_size
        except OBSException:
            raise OBSException(code=resp.errorCode, message=resp.errorMessage)
        except Exception as e:
            raise Exception("Get object size from OBS failed! ", e)

    @retry(
        wait=wait_fixed(1),
        stop=stop_after_attempt(3))
    def get_object(
            self,
            bucket_name,
            object_key,
            download_path=None,
            get_object_request=None,
            headers=None,
            load_stream_in_memory=False,
            progress_callback=None):
        """
        get object, support retry
        :param bucket_name:
        :param object_key:
        :param download_path:
        :param get_object_request:
        :param headers:
        :param load_stream_in_memory:
        :param progress_callback:
        :return:
        """
        return self.obs_client.getObject(
            bucketName=bucket_name,
            objectKey=object_key,
            downloadPath=download_path,
            getObjectRequest=get_object_request,
            headers=headers,
            loadStreamInMemory=load_stream_in_memory,
            progressCallback=progress_callback
        )

    @retry(
        wait=wait_fixed(1),
        stop=stop_after_attempt(3))
    def get_object_metadata(
            self,
            bucket_name,
            object_key,
            version_id=None,
            sse_header=None,
            origin=None,
            request_headers=None,
            extension_headers=None):
        """
        get object metadata, support retry
        :param bucket_name:
        :param object_key:
        :param version_id:
        :param sse_header:
        :param origin:
        :param request_headers:
        :param extension_headers:
        :return:
        """
        return self.obs_client.getObjectMetadata(
            bucketName=bucket_name,
            objectKey=object_key,
            versionId=version_id,
            sseHeader=sse_header,
            origin=origin,
            requestHeaders=request_headers,
            extensionHeaders=extension_headers)

    def delete_objects(self, bucket):
        """ Delete all objects in bucket
        return: Whether the deletion is successful
        """
        object_list = self.list_all_objects(bucket)
        for objects in object_list:
            delete_objects_resp = self.obs_client.deleteObject(bucket, objects)
            if delete_objects_resp.status > 300:
                print(delete_objects_resp.errorMessage)
                raise Exception("Delete files existed in bucket failed.")
        return True

    @retry(
        wait=wait_fixed(1),
        stop=stop_after_attempt(3))
    def delete_object(self, bucket_name, object_key, version_id=None, extension_headers=None):
        """
        delete obs object, support retry

        :param bucket_name:
        :param object_key:
        :param version_id:
        :param extension_headers:
        :return:
        """
        return self.obs_client.deleteObject(
            bucketName=bucket_name,
            objectKey=object_key,
            versionId=version_id,
            extensionHeaders=extension_headers)

    def create_directory(self, bucket, directory):
        """
        Create a directory in bucket.
        """
        if directory[-1] != "/":
            directory = directory + "/"

        if directory[0] == "/":
            directory = directory[1:]

        create_dir_resp = self.obs_client.putContent(bucket, directory,
                                                     content=None)
        if create_dir_resp.status < 300:
            print("Create directory %s successfully" % directory)
        else:
            raise Exception("Create directory failed. errorMessage {}".format(
                create_dir_resp.errorMessage))

    def list_all_objects(self, bucket, prefix=None):
        """
        list obs objects in bucket
        return: objects list[key, versionId]
        """
        objects_list = []
        obs_objects_lists = self._list_obs_objects(bucket, prefix)

        for list_objects_resp in obs_objects_lists:
            for content in list_objects_resp.body.contents:
                objects_list.append(content.key)

        return objects_list

    def _list_obs_objects(self, bucket, prefix=None):
        """
        list obs objects using maker in bucket
        return: obs response
        """
        obs_resp_list = []
        marker = None
        while True:
            list_objects_resp = self.obs_client.listObjects(bucketName=bucket,
                                                            prefix=prefix,
                                                            max_keys=100,
                                                            marker=marker)
            if list_objects_resp.status < 300:
                obs_resp_list.append(list_objects_resp)

                if not list_objects_resp.body.is_truncated:
                    break
                marker = list_objects_resp.body.next_marker
            else:
                obs_path = bucket if prefix is None else os.path.join(bucket, prefix)
                print("Your obs path {} does not exist.".format(obs_path))
                raise OBSException(code=list_objects_resp.errorCode,
                                   message=list_objects_resp.errorMessage)

        return obs_resp_list
